from collections import defaultdict, OrderedDict  # TODO: pryč s OrderedDict
import itertools
import re

from shlex import shlex


ATTRS = ['word', 'lemma'] + list('kegncpamdxytzw~')
ATTR_MAP = {attr: attr_id for (attr_id, attr) in enumerate(ATTRS)}
ATTR_IDS = list(range(len(ATTRS)))


# TODO: slovník sestavit až sekundárně (takže nedědit)
# TODO: podporovat append, del a přístup pod jménem/klíčem
class XmlTag(OrderedDict):
    def __init__(self, lines=None, tag=None, opening=None, empty=None):
        super().__init__()
        self.lines = lines
        self.tag = tag
        self.opening = opening
        self.empty = empty

    @property
    def closing(self):
        return not self.opening

    def add(self, attr, value):
        existing = self.get(attr, None)
        if existing is not None:
            if isinstance(existing, str):
                self[attr] = [existing, value]
            else:
                existing.append(value)
        else:
            self[attr] = value

    def __str__(self):
        opening = '' if self.opening else '/'
        empty = '/' if self.empty else ''
        # ztratil jsem původní pořadí duplikovaných atributů
        attrs = ''.join(
            ' {}="{}"'.format(attr, val) for attr, val in self.items())
        return '<{opening}{tag}{attrs}{empty}>'.format(
            opening=opening, tag=self.tag, attrs=attrs, empty=empty)

    def __repr__(self):
        return 'XmlTag(' + escape(self) + ')'

    @classmethod
    def parse(cls, lines):
        parser = shlex(lines[1:])
        xml = cls()
        attr = None
        value = None
        while True:
            token = parser.get_token()  # ValueError("No closing quotation")
            if not token:
                break
            elif xml.opening is None:
                xml.opening = token != '/'
                if xml.opening:
                    xml.tag = token
            elif xml.tag is None:
                if token == '>':
                    break
                xml.tag = token
            elif attr is None:
                if token == '/':
                    xml.empty = True
                elif token == '>':
                    break
                elif token == '=':
                    break  # unexpected =
                else:
                    attr = token
                    value = '='  # expecting =
            elif xml.empty:
                if token == '>':
                    break
                else:
                    break  # unexpected /
            elif value == '=':
                if token == '=':
                    value = None  # expecting a value
                elif token == '/':
                    xml.add(attr, None)
                    xml.empty = True
                elif token == '>':
                    xml.add(attr, None)
                    break
                else:
                    xml.add(attr, None)  # got an empty value
                    attr = token  # immediately followed by another
            elif value is None:
                if (len(token) > 1 and token[0] in '\'"' and
                        token[0] == token[-1]):
                    token = token[1:-1]
                xml.add(attr, token)
                attr = None  # expecting an attribute
                value = None
        return xml


def extract_attributes_from_tag(tag):
    # TODO: also return tag in .ambiguous so matching on tag works
    # MAYBE: support k1c14 (multiple-valued attributes in a short form)
    if ',' in tag:
        tags = tag.split(',')
        if '' not in tags:
            return put_ambiguous_attrs_aside(tags)
    if len(tag) % 2 != 0:
        raise ValueError('Malformed tag "{}"'.format(tag))
    return list(grouper(tag, 2)), []


def put_ambiguous_attrs_aside(tags):
    # MAYBE: support ambiguous tags of uneven length: c1,c2d1
    all_values = defaultdict(set)
    parsed_tags = []

    for tag in tags:
        if len(tag) % 2 != 0:
            raise ValueError('Malformed tag "{}"'.format(tag))
        attrs = list(grouper(tag, 2)) + [('tag', tag)]
        for attr, value in attrs:
            all_values[attr].add(value)
        parsed_tags.append(attrs)

    unambiguous = {attr: values.pop() for (attr, values) in all_values.items()
                   if len(values) == 1}
    ambiguous = []
    for attrs in parsed_tags:
        ambiguous_group = {attr: value for (attr, value) in attrs
                           if attr not in unambiguous}
        if ambiguous_group:
            ambiguous.append(ambiguous_group)
    return unambiguous, ambiguous


# copied from oVirt/VDSM, which in turn used
# http://docs.python.org/2.6/library/itertools.html?highlight=grouper#recipes
def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx
    args = [iter(iterable)] * n
    return itertools.zip_longest(fillvalue=fillvalue, *args)


need_escaping = re.compile(r'[()=? \t\n]').search  # podle shlex._find_unsafe


def escape(s):  # začátek podle shlex.quote
    if not isinstance(s, str):
        s = str(s)
    if not s:
        return "''"
    elif need_escaping(s) is None:
        return s
    elif "'" not in s:
        return "'" + s + "'"
    elif '"' not in s:
        return '"' + s + '"'
    else:
        raise ValueError("Can't escape, the string includes both quotes: " + s)
